#include "wifi_util.h"

#include "web_config.h"
#include "util.h"
#include "ntp.h"
#include "led_effects.h"
#include "sensor_console.h"

#if defined(ESP8266)
#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266httpUpdate.h>
#include <WiFiClientSecure.h>
#include <CertStoreBearSSL.h>
BearSSL::CertStore certStore;
#include <time.h>
#elif defined(ESP32)
#  include <WiFi.h>
#endif

const char *host = "transfer.hft-stuttgart.de";
const int httpsPort = 443;

// DigiCert High Assurance EV Root CA
const char trustRoot[] PROGMEM
    = R"EOF(
-----BEGIN CERTIFICATE-----
MIIDxTCCAq2gAwIBAgIQAqxcJmoLQJuPC3nyrkYldzANBgkqhkiG9w0BAQUFADBs
MQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3
d3cuZGlnaWNlcnQuY29tMSswKQYDVQQDEyJEaWdpQ2VydCBIaWdoIEFzc3VyYW5j
ZSBFViBSb290IENBMB4XDTA2MTExMDAwMDAwMFoXDTMxMTExMDAwMDAwMFowbDEL
MAkGA1UEBhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3
LmRpZ2ljZXJ0LmNvbTErMCkGA1UEAxMiRGlnaUNlcnQgSGlnaCBBc3N1cmFuY2Ug
RVYgUm9vdCBDQTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMbM5XPm
+9S75S0tMqbf5YE/yc0lSbZxKsPVlDRnogocsF9ppkCxxLeyj9CYpKlBWTrT3JTW
PNt0OKRKzE0lgvdKpVMSOO7zSW1xkX5jtqumX8OkhPhPYlG++MXs2ziS4wblCJEM
xChBVfvLWokVfnHoNb9Ncgk9vjo4UFt3MRuNs8ckRZqnrG0AFFoEt7oT61EKmEFB
Ik5lYYeBQVCmeVyJ3hlKV9Uu5l0cUyx+mM0aBhakaHPQNAQTXKFx01p8VdteZOE3
hzBWBOURtCmAEvF5OYiiAhF8J2a3iLd48soKqDirCmTCv2ZdlYTBoSUeh10aUAsg
EsxBu24LUTi4S8sCAwEAAaNjMGEwDgYDVR0PAQH/BAQDAgGGMA8GA1UdEwEB/wQF
MAMBAf8wHQYDVR0OBBYEFLE+w2kD+L9HAdSYJhoIAu9jZCvDMB8GA1UdIwQYMBaA
FLE+w2kD+L9HAdSYJhoIAu9jZCvDMA0GCSqGSIb3DQEBBQUAA4IBAQAcGgaX3Nec
nzyIZgYIVyHbIUf4KmeqvxgydkAQV8GK83rZEWWONfqe/EW1ntlMMUu4kehDLI6z
eM7b41N5cdblIZQB2lWHmiRk9opmzN6cN82oNLFpmyPInngiK3BD41VHMWEZ71jF
hS9OMPagMRYjyOfiZRYzy78aG6A9+MpeizGLYAiJLQwGXFK3xPkKmNEVX58Svnw2
Yzi9RKR/5CYrCsSXaQ3pjOLAEFe4yHYSkVXySGnYvCoCWw9E1CAx2/S6cCZdkGCe
vEsXCS+0yx5DaMkHJ8HSXPfqIbloEpw8nL+e/IBcm2PN7EeqJSdnoDfzAIJ9VNep
+OkuE6N36B9K
-----END CERTIFICATE-----
)EOF";
X509List cert(trustRoot);

extern const unsigned char caCert[] PROGMEM;
extern const unsigned int caCertLen;

namespace wifi {
  char local_ip[16]; // "255.255.255.255\0"

  void update_started() {
    Serial.println("CALLBACK:  HTTP update process started");
  }

  void update_finished() {
    Serial.println("CALLBACK:  HTTP update process finished");
  }

  void update_progress(int cur, int total) {
    Serial.printf("CALLBACK:  HTTP update process at %d of %d bytes...\n", cur, total);
  }

  void update_error(int err) {
    Serial.printf("CALLBACK:  HTTP update fatal error code %d\n", err);
  }

  void update() {
    ESPhttpUpdate.onStart(update_started);
    ESPhttpUpdate.onEnd(update_finished);
    ESPhttpUpdate.onProgress(update_progress);
    ESPhttpUpdate.onError(update_error);
    //TODO: Update https link to latest build
    WiFiClientSecure client;
    client.setTrustAnchors(&cert);
    if (!client.connect(host, httpsPort)) {
      Serial.println("Connection failed");
      return;
    }
    client.setInsecure();
    t_httpUpdate_return ret =
        ESPhttpUpdate.update(client,
            "https://transfer.hft-stuttgart.de/gitlab/co2ampel/ampel-firmware/-/jobs/13470/artifacts/raw/.pio/build/esp8266/firmware.bin");
    // Or:
    //t_httpUpdate_return ret = ESPhttpUpdate.update(client, "server", 80, "file.bin");
    //TODO: https://github.com/programmer131/ESP8266_ESP32_SelfUpdate

    switch (ret) {
    case HTTP_UPDATE_FAILED:
      Serial.printf("HTTP_UPDATE_FAILD Error (%d): %s\n", ESPhttpUpdate.getLastError(),
          ESPhttpUpdate.getLastErrorString().c_str());
      break;

    case HTTP_UPDATE_NO_UPDATES:
      Serial.println("HTTP_UPDATE_NO_UPDATES");
      break;

    case HTTP_UPDATE_OK:
      Serial.println("HTTP_UPDATE_OK");
      break;
    }
  }

  bool connected() {
    return WiFi.status() == WL_CONNECTED;
  }

  bool isAccessPoint() {
    return WiFi.getMode() == WIFI_AP;
  }

  /*
   * Connection attempt, called in blocking mode by setup(). This way, LED effects can be shown
   * without needing callbacks, but only during wifi_timeout seconds.
   * If connection fails, access point will be started indefinitely, and corresponding
   * LED effects will be shown during 5 seconds.
   *
   * Afterwards, the non-blocking web_config::update() will be called inside loop, and the ampel
   * can display CO2 levels.
   */
  void tryConnection() {
    for (int i = 0; i <= config::wifi_timeout + 5; i++) {
      web_config::update();
      sensor_console::checkSerialInput(); // To allow reset or ssid ... during startup
      if (isAccessPoint()) {
        led_effects::alert(0x1cff68);
      } else if (connected()) {
        break;
      } else {
        led_effects::showRainbowWheel();
      }
      Serial.print(".");
    }
    Serial.println();
  }

  void scanNetworks() {
    Serial.println();
    Serial.println(F("WiFi - Scanning..."));
    bool async = false;
    bool showHidden = true;
    int n = WiFi.scanNetworks(async, showHidden);
    for (int i = 0; i < n; ++i) {
      Serial.print(F("  * '"));
      Serial.print(WiFi.SSID(i));
      Serial.print(F("' ("));
      int16_t quality = 2 * (100 + WiFi.RSSI(i));
      Serial.print(util::min(util::max(quality, 0), 100));
      Serial.println(F(" %)"));
    }
    Serial.println(F("Done!"));
    Serial.println();
  }

  void showLocalIp() {
    Serial.print(F("WiFi - Local IP : "));
    Serial.println(wifi::local_ip);
    Serial.print(F("WiFi - SSID : "));
    Serial.println(config::selected_ssid());
  }

  void defineCommands() {
    sensor_console::defineCommand("wifi_scan", scanNetworks, F("(Scans available WiFi networks)"));
    sensor_console::defineCommand("local_ip", showLocalIp, F("(Displays local IP and current SSID)"));
    sensor_console::defineCommand("update!", update, F("(Updates to latest firmware)"));
    //TODO: Add "update!" command? https://github.com/esp8266/Arduino/blob/master/libraries/ESP8266httpUpdate/examples/httpUpdate/httpUpdate.ino
  }
}
